/*
 * This program generates an alignment check exception by doing an outsw/outsd
 * from an unaligned memory address.
 *
 * a.out -p 0xeee0              outsl buf, 0xeee0 
 * a.out -p 0xeee0 -o 1         outsl buf+1, 0xeee0     #AC
 * a.out -p 0xeee0 -w -o 2      outsw buf+2, 0xeee0
 * a.out -p 0xeee0 -w -o 3      outsw buf+3, 0xeee0     #AC
 * a.out -p 0xeee0 -o 3 -a      outsl buf+3, 0xeee0     
 * a.out -p 0xeee0 -o 1 -b      outsb buf+1, 0xeee0
 *
 * Test an outsl across a page boundary:
 * a.out -p 0xeee0 -a -o 4089 -c 3
 */

#include <sys/types.h>
#include <sys/sysctl.h>
#include <machine/cpufunc.h>
#include <machine/sysarch.h>
#include <x86/psl.h>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <libgen.h>

static char buf[8192] __aligned(4096);
static char *progname;

static void
usage(int rc)
{
	fprintf(stderr,
		"Usage: %s [-h][-a][-bwl][-c <count>][-o <offset>] -p <port>\n"
		"       -h: display usage\n"
		"       -a: disable alignment check\n"
		"       -c: count (default is 1)\n"
		"       -b: outsb\n"
		"       -w: outsw\n"
		"       -l: outsl (default)\n"
		"       -o: offset (default is 0)\n"
		"       -p: port number (4 bytes wide)\n",
		progname);
	exit(rc);
}

int
main(int argc, char *argv[])
{
	struct i386_ioperm_args ioperm;
	long rflags;
	char *ptr;
	const char *str = "deadbeef";
	int ch, count, error, enable_ac, i, port, size, offset;

	progname = basename(argv[0]);

	count = 1;
	port = -1;
	offset = 0;
	size = 4;
	enable_ac = 1;

	while ((ch = getopt(argc, argv, "abwlhc:o:p:")) != -1) {
		switch (ch) {
		case 'a':
			enable_ac = 0;
			break;
		case 'c':
			count = strtol(optarg, NULL, 0);
			break;
		case 'b':
			size = 1;
			break;
		case 'w':
			size = 2;
			break;
		case 'l':
			size = 4;
			break;
		case 'p':
			port = strtol(optarg, NULL, 0);
			break;
		case 'o':
			offset = atoi(optarg);
			break;
		case 'h':
			usage(0);
			break;
		default:
			usage(1);
			break;
		}
	}

	if (port < 0 || port >= 65536)
		usage(1);

	if (offset < 0)
		usage(1);

	if (count <= 0)
		usage(1);

	i = 0;
	ptr = buf;
	while (ptr < buf + sizeof(buf)) {
		*ptr++ = str[i++];
		if (str[i] == '\0')
			i = 0;
	}

	i = 1;
	sysctlbyname("machdep.uprintf_signal", NULL, NULL, &i, sizeof(int));

	ioperm.start = port;
	ioperm.length = 4;
	ioperm.enable = 1;
	if (sysarch(I386_SET_IOPERM, &ioperm) != 0) {
		perror("ioperm");
		exit(1);
	}

	if (enable_ac) {
		rflags = read_rflags();
		write_rflags(rflags | PSL_AC);
	}

	ptr = buf + offset;
	if (size == 1)
		outsb(port, ptr, count);
	else if (size == 2)
		outsw(port, ptr, count);
	else
		outsl(port, ptr, count);

	rflags = read_rflags();
	write_rflags(rflags & ~PSL_AC);

	exit(0);
}
