/*
 * This program triggers a #GP by using a non-canonical source address for the
 * 'outsb' instruction. It does this by using 0x7ffffffffff0 as the source
 * address.
 *
 * The address 0x800000000000 is non-canonical will cause the cpu to trigger
 * a protection fault.
 *
 * ./a.out -p 0xeee0		outsb(0x7fffffffffff, 0xeee0, 1)
 * ./a.out -p 0xeee0 -c 16 	outsb(0x7fffffffffff, 0xeee0, 16)
 * ./a.out -p 0xeee0 -c 17 	outsb(0x7fffffffffff, 0xeee0, 17)	#GP
 */
#include <sys/param.h>
#include <sys/sysctl.h>

#include <vm/pmap.h>

#include <machine/cpufunc.h>
#include <machine/pmap.h>
#include <machine/vmparam.h>
#include <machine/sysarch.h>

#include <stdio.h>
#include <stdlib.h>
#include <libgen.h>
#include <unistd.h>

#ifndef CTASSERT
#define	CTASSERT(x)	_Static_assert(x, "compile-time assertion failed")
#endif

#ifndef __amd64__
#error "must be compiled on amd64 platforms"
#endif

CTASSERT(VM_MAXUSER_ADDRESS == 0x800000000000);

static char *progname;

static void
usage(int rc)
{
	fprintf(stderr,
		"Usage: %s [-h][-a <address>][-c <count>] -p <port>\n"
		"       -a: source address (default is %#lx)\n"
		"       -c: count of bytes written to i/o port (default is 1)\n"
		"       -h: display usage\n"
		"       -p: port number\n",
		progname, VM_MAXUSER_ADDRESS - 1);
	exit(rc);
}

int
main(int argc, char *argv[])
{
	struct i386_ioperm_args ioperm;
	int ch, count, i, port;
	u_long srcaddr;

	count = 1;
	port = -1;
	srcaddr = VM_MAXUSER_ADDRESS - 16;
	progname = basename(argv[0]);

	while ((ch = getopt(argc, argv, "ha:c:p:")) != -1) {
		switch (ch) {
		case 'a':
			srcaddr = strtoul(optarg, NULL, 0);
			break;
		case 'c':
			count = strtol(optarg, NULL, 0);
			break;
		case 'p':
			port = strtol(optarg, NULL, 0);
			break;
		case 'h':
			usage(0);
		default:
			usage(1);
		}
	}

	if (port < 0 || port >= 65536 || count <= 0)
		usage(1);

	i = 1;
	sysctlbyname("machdep.uprintf_signal", NULL, NULL, &i, sizeof(int));

	ioperm.start = port;
	ioperm.length = 4;
	ioperm.enable = 1;
	if (sysarch(I386_SET_IOPERM, &ioperm) != 0) {
		perror("ioperm");
		exit(1);
	}

	outsb(port, (void *)(uintptr_t)srcaddr, count);
	printf("Successfully wrote %d bytes from %#lx to i/o port %d\n",
	    count, srcaddr, port);
}
