/*
 * ./a.out -p 0xeee0
 * ./a.out -p 0xeee1
 * ./a.out -p 0xeee2
 * ./a.out -p 0xeee3
 * ./a.out -p 0xeee0 -c 4 -o 4090 -w
 * ./a.out -p 0xeee0 -c 4 -o 4089		#AC
 * ./a.out -p 0xeee0 -c 4 -o 4089 -a
 */

#include <sys/types.h>
#include <sys/sysctl.h>
#include <machine/cpufunc.h>
#include <machine/sysarch.h>
#include <x86/psl.h>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <libgen.h>
#include <string.h>

static char buf[8192] __aligned(4096);
static char buf2[8192] __aligned(4096);
static char *progname;

static void
usage(int rc)
{
	fprintf(stderr,
		"Usage: %s [-h][-a][-bwl][-c <count>][-o <offset>] -p <port>\n"
		"       -h: display usage\n"
		"       -a: disable alignment check\n"
		"       -c: count (default is 1)\n"
		"       -b: insb\n"
		"       -w: insw\n"
		"       -l: insl (default)\n"
		"       -o: offset (default is 0)\n"
		"       -p: port number (4 bytes wide)\n",
		progname);
	exit(rc);
}

int
main(int argc, char *argv[])
{
	struct i386_ioperm_args ioperm;
	long rflags;
	char *ptr;
	const char *str = "deadbeef";
	int ch, count, error, enable_ac, i, port, size, offset;
	char p[4];

	progname = basename(argv[0]);

	count = 1;
	port = -1;
	offset = 0;
	size = 4;
	enable_ac = 1;

	while ((ch = getopt(argc, argv, "abwlhc:o:p:")) != -1) {
		switch (ch) {
		case 'a':
			enable_ac = 0;
			break;
		case 'c':
			count = strtol(optarg, NULL, 0);
			break;
		case 'b':
			size = 1;
			break;
		case 'w':
			size = 2;
			break;
		case 'l':
			size = 4;
			break;
		case 'p':
			port = strtol(optarg, NULL, 0);
			break;
		case 'o':
			offset = atoi(optarg);
			break;
		case 'h':
			usage(0);
			break;
		default:
			usage(1);
			break;
		}
	}

	if (port < 0 || port >= 65536)
		usage(1);

	if (offset < 0)
		usage(1);

	if (count <= 0)
		usage(1);

	i = 1;
	sysctlbyname("machdep.uprintf_signal", NULL, NULL, &i, sizeof(int));

	ioperm.start = port;
	ioperm.length = 4;
	ioperm.enable = 1;
	if (sysarch(I386_SET_IOPERM, &ioperm) != 0) {
		perror("ioperm");
		exit(1);
	}

	/*
	 * First, figure out the pattern being returned by the guest.
	 */
	for (i = 0; i < 4; i++)
		insb(port + i, p + i, 1);

	printf("pattern: %c%c%c%c\n", p[0], p[1], p[2], p[3]);

	for (i = 0; i < count; i++)
		bcopy(p, buf2 + i * size, size);

	if (enable_ac) {
		rflags = read_rflags();
		write_rflags(rflags | PSL_AC);
	}

	ptr = buf + offset;
	if (size == 1)
		insb(port, ptr, count);
	else if (size == 2)
		insw(port, ptr, count);
	else
		insl(port, ptr, count);

	rflags = read_rflags();
	write_rflags(rflags & ~PSL_AC);

	if (memcmp(ptr, buf2, count * size) == 0) {
		printf("Success\n");
		exit(0);
	} else {
		printf("Failed!!!\n");
		exit(1);
	}
}
